/*
    SmartHomeGeeks.io Smart Pet Feeder
    https://smarthomegeeks.io/smart-pet-feeder
*/

#include <FS.h>
#include <WiFiManager.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <ArduinoJson.h>
#include <PubSubClient.h>
#include <ArduinoOTA.h>

// Device configuration
const String deviceName = "Pet Feeder";

const int motor = 12;       // pull HIGH to turn on
const int switch1 = 13;     // LOW to HIGH equals one portion
const int playbutton = 14;  // pull HIGH to play recorded message
const int ledbutton = A0;   // HIGH when pressed
const int feedbutton = 5;   // HIGH when pressed

char mqtt_server[40];
char mqtt_port[6] = "1883";
char mqtt_user[96];
char mqtt_pass[96];
char mqtt_client[40] = "pet_feeder";
bool playRecording = true;
int portionSize = 1;

unsigned long lastMQTTconnect = 0;

WiFiClient espClient;
PubSubClient client(espClient);

//flag for saving data
bool shouldSaveConfig = false;

//callback notifying us of the need to save config
void saveConfigCallback () {
  Serial.println("Should save config");
  shouldSaveConfig = true;
}

void autoDiscovery() {
  Serial.println("Publishing Home Assistant Autodiscovery...");

  char buffer[1024];

  DynamicJsonDocument deviceData(512);
  deviceData["name"] = deviceName;
  deviceData["ids"] = mqtt_client;
  deviceData["mdl"] = "Pet Feeder";
  deviceData["mf"] = "SmartHomeGeeks.io";
  deviceData["sw"] = "2023.10";
  
  // Switch
  DynamicJsonDocument root(1024);
  root["name"] = "Recording";
  root["obj_id"] = String(mqtt_client) + "_play";
  root["ic"] = "mdi:account-voice";
  root["uniq_id"] = "switch." + String(mqtt_client);
  root["stat_t"] = String(mqtt_client) + "/playStatus";
  root["cmd_t"] = String(mqtt_client) + "/playCmd";
  root["avty_t"] = String(mqtt_client) + "/availability";
  root["ent_cat"] = "config";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/switch/" + String(mqtt_client) + "/config").c_str(), buffer, true);
  root.clear();

  // Feed Button 
  root["name"] = "Single Portion";
  root["obj_id"] = String(mqtt_client) + "_feedbutton";
  root["ic"] = "mdi:numeric-1";
  root["uniq_id"] = "button." + String(mqtt_client) + "_feedbutton";
  root["cmd_t"] = String(mqtt_client) + "/feedbuttonCmd";
  root["avty_t"] = String(mqtt_client) + "/availability";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/button/" + String(mqtt_client) + "/feed/config").c_str(), buffer, true);
  root.clear();

  // LED Button 
  root["name"] = "Dispense Meal";
  root["obj_id"] = String(mqtt_client) + "_ledbutton";
  root["ic"] = "mdi:counter";
  root["uniq_id"] = "button." + String(mqtt_client) + "_ledbutton";
  root["cmd_t"] = String(mqtt_client) + "/ledbuttonCmd";
  root["stat_t"] = String(mqtt_client) + "/ledbuttonStatus";
  root["avty_t"] = String(mqtt_client) + "/availability";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/button/" + String(mqtt_client) + "/led/config").c_str(), buffer, true);
  root.clear();

  // Play Button 
  root["name"] = "Play Recording";
  root["obj_id"] = String(mqtt_client) + "_playbutton";
  root["ic"] = "mdi:chat";
  root["uniq_id"] = "button." + String(mqtt_client) + "_playbutton";
  root["cmd_t"] = String(mqtt_client) + "/playbuttonCmd";
  root["avty_t"] = String(mqtt_client) + "/availability";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/button/" + String(mqtt_client) + "/play/config").c_str(), buffer, true);
  root.clear();

  // Input for portion size
  root["name"] = "Portion Size";
  root["obj_id"] = String(mqtt_client) + "_portionSize";
  root["ic"] = "mdi:scale";
  root["uniq_id"] = "number." + String(mqtt_client) + "_portion_size";
  root["cmd_t"] = String(mqtt_client) + "/portion_size";
  root["stat_t"] = String(mqtt_client) + "/portion_size";
  root["min"] = 1;
  root["max"] = 10;
  root["step"] = 1;
  root["mode"] = "slider";
  root["avty_t"] = String(mqtt_client) + "/availability";
  root["ent_cat"] = "config";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/number/" + String(mqtt_client) + "/portion_size/config").c_str(), buffer, true);
  root.clear();

  // status sensor
  root["name"] = nullptr;
  root["obj_id"] = String(mqtt_client) + "_status";
  root["stat_t"] = String(mqtt_client) + "/status";
  root["ic"] = "mdi:bowl";
  root["uniq_id"] = "sensor." + String(mqtt_client);
  root["avty_t"] = String(mqtt_client) + "/availability";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqtt_client) + "/config").c_str(), buffer, true);
  root.clear();

  // RSSI sensor
  root["name"] = "RSSI";
  root["obj_id"] = String(mqtt_client) + "_rssi";
  root["stat_t"] = String(mqtt_client) + "/rssi";
  root["unit_of_meas"] = "dBm";
  root["ic"] = "mdi:access-point";
  root["uniq_id"] = "sensor." + String(mqtt_client) + "_rssi";
  root["avty_t"] = String(mqtt_client) + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqtt_client) + "/rssi/config").c_str(), buffer, true);
  root.clear();

  // IP
  root["name"] = "IP";
  root["obj_id"] = String(mqtt_client) + "_ip";
  root["stat_t"] = String(mqtt_client) + "/ip";
  root["ic"] = "mdi:ip-network";
  root["uniq_id"] = "sensor." + String(mqtt_client) + "_ip";
  root["avty_t"] = String(mqtt_client) + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqtt_client) + "/ip/config").c_str(), buffer, true);
  root.clear();

  client.publish((String(mqtt_client) + "/playStatus").c_str(), (playRecording ? "ON" : "OFF"), true);
  client.publish((String(mqtt_client) + "/portion_size").c_str(), String(portionSize).c_str(), true);
  client.publish((String(mqtt_client) + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);

}

void connectmqtt() {
  // prevent constant reconnect
  if (millis() - lastMQTTconnect < 5000) return;
  lastMQTTconnect = millis();

  if (client.connect(mqtt_client, mqtt_user, mqtt_pass, (String(mqtt_client) + "/availability").c_str(), 1, true, "offline")) {
    Serial.println("connected to MQTT");
    client.setBufferSize(512);
    autoDiscovery();

    client.subscribe((String(mqtt_client) + "/playbuttonCmd").c_str());
    client.subscribe((String(mqtt_client) + "/ledbuttonCmd").c_str());
    client.subscribe((String(mqtt_client) + "/feedbuttonCmd").c_str());
    client.subscribe((String(mqtt_client) + "/playCmd").c_str());
    client.subscribe((String(mqtt_client) + "/portion_size").c_str());
    client.publish((String(mqtt_client) + "/availability").c_str(), "online", true);
    client.publish((String(mqtt_client) + "/ip").c_str(),  WiFi.localIP().toString().c_str());

  } else {
    Serial.print("failed, rc=");
    Serial.print(client.state());
    Serial.println(" try again in 5 seconds");
  }
}

void setup() {
  // put your setup code here, to run once:
  Serial.begin(115200);
  delay(1000);
  Serial.println();

  pinMode(switch1, INPUT_PULLUP);
  pinMode(feedbutton, INPUT);
  pinMode(motor, OUTPUT);
  pinMode(playbutton, OUTPUT);
  digitalWrite(motor, LOW);

  //read configuration from FS json
  Serial.println("mounting FS...");

  if (SPIFFS.begin()) {
    Serial.println("mounted file system");
    if (SPIFFS.exists("/config.json")) {
      //file exists, reading and loading
      Serial.println("reading config file");
      File configFile = SPIFFS.open("/config.json", "r");
      if (configFile) {
        Serial.println("opened config file");
        size_t size = configFile.size();
        // Allocate a buffer to store contents of the file.
        std::unique_ptr<char[]> buf(new char[size]);

        configFile.readBytes(buf.get(), size);

        DynamicJsonDocument json(1024);
        auto deserializeError = deserializeJson(json, buf.get());
        serializeJson(json, Serial);
        if ( ! deserializeError ) {
          Serial.println("\nparsed json");
          strcpy(mqtt_server, json["mqtt_server"]);
          strcpy(mqtt_port, json["mqtt_port"]);
          strcpy(mqtt_user, json["mqtt_user"]);
          strcpy(mqtt_pass, json["mqtt_pass"]);
          strcpy(mqtt_client, json["mqtt_client"]);
          playRecording = json["playRecording"];
          portionSize = (json["portionSize"] == nullptr ? 1 : json["portionSize"]);
          client.publish((String(mqtt_client) + "/playStatus").c_str(), (playRecording ? "ON" : "OFF"), true);
        } else {
          Serial.println("failed to load json config");
        }
        configFile.close();
      }
    }
  } else {
    Serial.println("failed to mount FS");
  }
  //end read

  // The extra parameters to be configured (can be either global or just in the setup)
  // After connecting, parameter.getValue() will get you the configured value
  // id/name placeholder/prompt default length
  WiFiManagerParameter custom_mqtt_server("mqtt_server", "MQTT server", mqtt_server, 40);
  WiFiManagerParameter custom_mqtt_port("mqtt_port", "MQTT port", mqtt_port, 6);
  WiFiManagerParameter custom_mqtt_user("mqtt_user", "MQTT Username", mqtt_user, 64);
  WiFiManagerParameter custom_mqtt_pass("mqtt_pass", "MQTT Password", mqtt_pass, 64);
  WiFiManagerParameter custom_mqtt_client("mqtt_client", "MQTT Topic", mqtt_client, 40);

  //WiFiManager
  //Local intialization. Once its business is done, there is no need to keep it around
  WiFiManager wifiManager;

  // Set WiFi Hostname
  WiFi.hostname(mqtt_client);

  //set config save notify callback
  wifiManager.setSaveConfigCallback(saveConfigCallback);

  //add all your parameters here
  wifiManager.addParameter(&custom_mqtt_server);
  wifiManager.addParameter(&custom_mqtt_port);
  wifiManager.addParameter(&custom_mqtt_user);
  wifiManager.addParameter(&custom_mqtt_pass);
  wifiManager.addParameter(&custom_mqtt_client);

  // configure a portal timeout and restart
  wifiManager.setConfigPortalTimeout(300);

  if (!wifiManager.autoConnect("SHGio_Pet_Feeder")) {
    Serial.println("failed to connect and hit timeout");
    delay(5000);
    ESP.restart();
  }

  //if you get here you have connected to the WiFi
  Serial.println("connected...yeey :)");

  //read updated parameters
  strcpy(mqtt_server, custom_mqtt_server.getValue());
  strcpy(mqtt_port, custom_mqtt_port.getValue());
  strcpy(mqtt_user, custom_mqtt_user.getValue());
  strcpy(mqtt_pass, custom_mqtt_pass.getValue());
  strcpy(mqtt_client, custom_mqtt_client.getValue());
  Serial.println("The values in the file are: ");
  Serial.println("\tmqtt_server : " + String(mqtt_server));
  Serial.println("\tmqtt_port : " + String(mqtt_port));
  Serial.println("\tmqtt_user : " + String(mqtt_user));
  Serial.println("\tmqtt_pass : " + String(mqtt_pass));
  Serial.println("\tmqtt_client : " + String(mqtt_client));
  Serial.println("\tplayRecording : " + String(playRecording));
  Serial.println("\tportionSize : " + String(portionSize));

  if (shouldSaveConfig) saveConfig();

  Serial.println("local ip");
  Serial.println(WiFi.localIP());

  // Configure OTA
  ArduinoOTA.setHostname(mqtt_client);
  ArduinoOTA.begin(); 

  // Connect MQTT
  client.setServer(mqtt_server, atoi(mqtt_port));
  client.setCallback(callback);
}

void saveConfig() {
  //save the custom parameters to FS
    Serial.println("saving config");
    DynamicJsonDocument json(1024);
    json["mqtt_server"] = mqtt_server;
    json["mqtt_port"] = mqtt_port;
    json["mqtt_user"] = mqtt_user;
    json["mqtt_pass"] = mqtt_pass;
    json["mqtt_client"] = mqtt_client;
    json["playRecording"] = playRecording;
    json["portionSize"] = portionSize;

    File configFile = SPIFFS.open("/config.json", "w");
    if (!configFile) {
      Serial.println("failed to open config file for writing");
    }

    serializeJson(json, Serial);
    serializeJson(json, configFile);
    configFile.close();
    //end save
}

void callback(char* topic, byte* payload, unsigned int length) {
  Serial.print("Topic arrived: ");
  Serial.println(String(topic));
  
  Serial.print("Message arrived: ");
  for (int i=0;i<length;i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();

  if (String(topic) == String(mqtt_client) + "/playCmd") {
    playRecording = (!strncmp((char*)payload, "ON", length) ? true : false);
    client.publish((String(mqtt_client) + "/playStatus").c_str(), (!strncmp((char*)payload, "ON", length) ? "ON" : "OFF"), true);
    Serial.print("Changed playRecording value to: ");
    Serial.println(String(playRecording));
    saveConfig();
  } else if (String(topic) == String(mqtt_client) + "/ledbuttonCmd")  {
    feed(portionSize);
  } else if (String(topic) == String(mqtt_client) + "/feedbuttonCmd")  {
    feed(1);
  } else if (String(topic) == String(mqtt_client) + "/playbuttonCmd")  {
    play();
  } else if (String(topic) == String(mqtt_client) + "/portion_size")  {
    payload[length] = '\0';
    String s = String((char*)payload);
    portionSize = s.toInt();
    saveConfig();
  }

  client.publish((String(mqtt_client) + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
}

void play() {
  digitalWrite(playbutton, HIGH);
  delay(100);
  digitalWrite(playbutton, LOW);
}

void feed(int portions) {
  Serial.println("Start Feeding");

  client.publish((String(mqtt_client) + "/status").c_str(), "Feeding", true);

  Serial.print("portions = ");
  Serial.println(portions);

  int count = 0;
  bool prevState = digitalRead(switch1); 
  bool newState = digitalRead(switch1); 

  // play recorded message
  if (playRecording) {
    play();
    // added a delay because the noise from the food was making it hard to hear the recording.
    delay(3000);
  }

  // turn the motor on
  Serial.println("Motor on");
  digitalWrite(motor, HIGH);

  unsigned long startMotor = millis();   

  Serial.print("count = ");
  Serial.println(count);

  // if switch state changes from OFF to ON, increase count
  while (count < portions && millis() - startMotor < 5000) {
      newState = digitalRead(switch1);
    	if (prevState != newState && !newState) {
        startMotor = millis();
        count++;
        Serial.print("Increase portion count to: ");
        Serial.println(count);
      }
      prevState = newState;
      delay(50);
  }

  if (count != portions) {
    client.publish((String(mqtt_client) + "/status").c_str(),  "Error", true);
  } else {
    client.publish((String(mqtt_client) + "/status").c_str(),  "Idle", true);
  }

  // turn the motor off
  Serial.println("Motor off");
  digitalWrite(motor, LOW);
}

void loop() {
  if (!client.connected()) connectmqtt();

  ArduinoOTA.handle();

  if (digitalRead(feedbutton) && analogRead(ledbutton) > 100) {
    delay(5000);
    if (digitalRead(feedbutton) && analogRead(ledbutton) > 100) {
      // reset WiFi settings
     WiFiManager wifiManager;
     wifiManager.resetSettings();
      SPIFFS.format();
      ESP.restart();
    }
  } else if (digitalRead(feedbutton)) {
    while (digitalRead(feedbutton)) delay(50);
    int feedCount = 1;
    unsigned long startFeedButton = millis();   
    while (millis() - startFeedButton < 1000) {
      while (digitalRead(feedbutton)) delay(50);
      if (digitalRead(feedbutton)) {
        startFeedButton = millis();
        feedCount++;
      }
    }
    feed(feedCount);
  } else if (analogRead(ledbutton) > 100) {
    client.publish((String(mqtt_client) + "/ledbuttonStatus").c_str(), "ON");
    while (analogRead(ledbutton) > 100) delay(100);
    client.publish((String(mqtt_client) + "/ledbuttonStatus").c_str(), "OFF");
    feed(portionSize);
  }

  client.loop();

  delay(100);

}